/*
This code demonstrates a macro for creating a data dictionary based on SAS variable names, labels, and formats.

In my role as a data coordinating center statistician, I'm often asked to share limited datasets (i.e., limited to certain subjects and limited to certain analysis variables) from 
standard analysis files with non-SAS users. If were sharing with someone in SAS, I could likely forego the data dictionary and just label the variables and give them a proc formats
statement for any coded numeric values. However, these are more often going to the study PI who likely doesn't use SAS, likely isn't a statistician, and generally prefers a .csv or
excel file that they can easily open and review--before bringing the data to their own statistician. So when I'm creating these "bespoke" files, I've *already* had to export the raw data
from the database, create analysis variables, add variable labels/formats etc. and save as the standard analysis file. And then I have to limit the data to the participants and variables
they have access to. The purpose of this macro is to automate the data dictionary creation after completing those steps.


The first section will create a small example dataset; then the actual macro is defined and used.

Jonathan Troost
MICHR
01/11/2021
troostj@med.umich.edu
*/

/*First, we will create a small dataset with labels and formats. When using your own data, just make sure to save the necessary formats to a dataset using "cntlout=" in proc format*/
proc format cntlout=formats;
	value Sex
	1="Female"
	2="Male";
	value Race
	1="American Indian/Alaska Native" 
	2="Asian"
	3="Native Hawaiian or Other Pacific Islander" 
	4="Black or African American" 
	5="White"
	6="More Than One Race" 
	7="Unknown / Not Reported";
run;
data Example;
	length ID $7.;
	input ID Albumin Age Sex Race ;
	label ID="Study ID" Albumin="Serum Albumin (g/dL)" Age="Age (years)" Sex="Sex" Race="Race";
	format Sex Sex. Race Race.;
	cards;
001-001 4.4 26 1 5
001-002 4.9 54 2 4
001-003 2.4 38 2 4
001-004 3.1 26 1 5
001-005 2.9 16 2 2
002-001 3.2 29 1 5
002-002 3.7 49 2 6
002-003 3.3 54 1 5
002-004 3.1 55 1 5
002-005 1.9 40 2 4
;
run;
/*In reality, you might already your format catalogue saved in a library somewhere and there's an easier way to create the "formats" dataset than as part of the proc format statement
as shown above. If that's the case, try this:

proc format cntlout=formats library=<LIBNAME where formats catalogue is saved>;
run;

However, this approach requires that the formats catalogue was made in the same operating system as you are currently using--so it's possible you can run into 32-bit vs. 64-bit windows
issues. I would recommend whenever creatinine a permanent formats catalogue to also create a formats dataset with cntlout--which will make it easier to transfer.
*/


/*The macro itself just needs a SAS dataset, the format dataset, and the name of the output dataset for the data dictionary*/
%macro makeDD(data=, formatdata=, dd=, maxlength=4000);

/*Most of the meta data we want comes out of proc contents. It will output a dataset where each row is a variable is a row, and there are variables for the var name, label, type,
length etc. It won't quite get us what we need for the coded-numeric formats. But it will tell us the format name which is a start*/
proc contents data=&data out=_contents noprint;run;
data _contents2;
	set _contents;
	FMTNAME=FORMAT;
run;

/*Now we're going to create a flat dataset for each format name where there's a single large cell with the entire format decoded. So while the raw formats dataset will have a row
that says 1="Female" and another saying 2="Male", we will get a single cell that says 1="Female" | 2="Male" to put in our data dictionary*/
data _formats;
	set &formatdata;
run;
proc sort data=_formats; by FMTNAME Start;run;
data _formats2;
	set _formats;
	retain valuen 0;
	by FMTNAME;
	if first.FMTNAME then valuen=1;
	else valuen=valuen+1;
	Value=strip(START)||"="||LABEL;
run;
/*This part is counting up the maximum number of levels within a format to it knows how to properly concatenate them later*/
proc means data=_formats2 max noprint;
	var valuen;
	output out=max max=max;
run;
data _null_;
	set max;
	call symputx("Max",max);
run;
proc transpose data=_formats2 out=_formats3;
	by FMTNAME;
	id valuen;
	var Value;
run;
/*Just make sure the maximum length for "format" is long enough to fit the entire pipe-delimited decode. The default sets this to $4000. which should be plenty but it's an optional
macro variable if you need more*/
data _formats4;
	length format $&maxlength..;
	set _formats3;
	Format=catx(" | ", of _1-_&Max);
run;
/*This sections brings everything together*/
proc sql;
	create table _dd as
	select a.MEMNAME as Table, a.VARNUM, a.NAME, a.Label, a.Type, A.Length, B.Format
	from _contents2 as a left join _formats4 as b
	on a.FMTNAME=b.FMTNAME;
quit;
proc sort data=_dd ; by VARNUM;run;
data &dd;
	retain Table VARNUM NAME Label VarType Length Format;
	length VarType $40.;
	set _dd;
	if Type=2 then VarType="Character";
	if Type=1 then VarType="Numeric";
	
	label Table="Dataset" format="Variable Format" VarType="Variable Type";
	keep Table VARNUM NAME Label VarType Length Format;
run;


%mend;
%makeDD(data=Example, formatdata=formats, dd=Example_DD);



/*Now we have a data dictionary for this dataset showing the dataset name, the variable name, order, label, type (character vs. numeric), length, and de-code for any format.
Since we keep the dataset name, we can append together data dictionaries for multiple datasets if needed as well.
*/
proc print data=Example noobs label;
run;

proc print data=Example_DD noobs label;
run;

/*
Suggest adding this to the body of the macro if you'd like to remove the intermediary datasets
proc datasets library=work;
	delete _:;
quit;
*/
